#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <time.h>

struct request {
	uint8_t command;
	uint8_t imageid;
	uint32_t offset;
	uint32_t length;
} __attribute__((packed));

enum {
	sCmdSize = 1,
	sCmdRead,
	sCmdWrite,
	sCmdEject
};

struct response {
	uint8_t status;
	uint8_t imageid;
	uint32_t length;
} __attribute__((packed));

void usage(const char *progname) {
	fprintf(stderr, "%s Options:\n", progname);
	fprintf(stderr, "\t-d <device>\tAttach the server to the given serial device.\n");
	fprintf(stderr, "\t--stdio\tVend the server over stdin/stdout, for use with emulators.\n");
	fprintf(stderr, "\t--logfile <file>\tPath to a server log file.\n");
	fprintf(stderr, "\t--disk <file>\tPath to the disk image to be shared over serial.\n");
	fprintf(stderr, "\t--baud [9600,19200,38400,57600]\tOne of the listed baud rates.\n");
	return;
}

/* Returns a stack allocated string for logging.  */
char *serlogtime() {
	static char timestr[1024];
	struct tm *t = NULL;
	time_t now = time(NULL);

	t = localtime(&now);
	if(!t) {
		strlcpy(timestr, "nulltime", sizeof(timestr));
		return timestr;
	}

	strftime(timestr, sizeof(timestr), "%Y-%m-%d-%H:%M:%S", t);

	return timestr;
}

int main(int argc, char *argv[]) {
	const char *progname = argv[0];

	int ifd, ofd;
	int serfd;
	struct stat sb;
	FILE *logfile = stderr;
	speed_t baud = B57600;

	/* options descriptor */
	bool use_stdio = false;
	char *device = NULL;
	char *logfile_path = NULL;
	char *image_path = NULL;
	static struct option longopts[] = {
		{ "stdio",		no_argument,		NULL,	's'	},
		{ "device",		required_argument,	NULL,	'd'	},
		{ "logfile",	required_argument,	NULL,	'l'	},
		{ "disk",		required_argument,	NULL,	'i'	},
		{ "baud",		required_argument,	NULL,	'b'	},
		{ NULL,			0,					NULL,	0	}
	};

	int ch;
	while ((ch = getopt_long(argc, argv, "b:sd:li:", longopts, NULL)) != -1) {
		switch (ch) {
			case 'b':
				baud = (speed_t)strtol(optarg, NULL, 10);
				break;
			case 's':
				use_stdio = true;
				break;
			case 'd':
				device = optarg;
				break;
			case 'l':
				logfile_path = optarg;
				break;
			case 'i':
				image_path = optarg;
				break;
			default:
				usage(progname);
		}
	}
	argc -= optind;
	argv += optind;

	if ((!device && !use_stdio) || (device && use_stdio)) {
		usage(progname);
		exit(1);
	}

	if (image_path == NULL) {
		usage(progname);
		exit(1);
	}

	if (logfile_path) {
		logfile = fopen(logfile_path, "a");
		if (logfile == NULL) {
			fprintf(stderr, "Could not open log file %s: %d %s\n", logfile_path, errno, strerror(errno));
			exit(1);
		}

		setvbuf(logfile, (char *) NULL, _IONBF, 0);
	}

	if (device) {
		int fd = open(device, O_RDWR | O_NOCTTY );
		if(fd < 0) {
			fprintf(logfile, "Could not open device %s: %d %s\n", device, errno, strerror(errno));
			exit(1);
		}

		/* Same input/output fd for an actual serial device */
		ifd = fd;
		ofd = fd;

		struct termios opts;
		tcgetattr(fd, &opts);
		cfsetispeed(&opts, baud);
		cfsetospeed(&opts, baud);

		/* 8N1 */
		opts.c_cflag &= ~PARENB;
		opts.c_cflag &= ~CSTOPB;
		opts.c_cflag &= ~CSIZE;
		opts.c_cflag |= CS8;

		/* Disable hw flow control */
		opts.c_cflag &= ~CRTSCTS;

	#if 0
		/* Software flow control */
		// opts.c_iflag |= (IXON | IXOFF | IXANY);

		/* raw input */
		//opts.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

		/* raw output */
		//opts.c_oflag &= ~OPOST;
	#endif

		cfmakeraw(&opts);

		tcsetattr(fd, TCSANOW, &opts);
	} else {
		ifd = STDIN_FILENO;
		ofd = STDOUT_FILENO;
	}

	serfd = open(image_path, O_RDWR);
	if (serfd < 0) {
		fprintf(logfile, "Could not open disk image %s: %d %s\n", image_path, errno, strerror(errno));
		exit(1);
	}
	fstat(serfd, &sb);

	fprintf(logfile, "Server running ...\n");
	do {
		struct request req;
		uint8_t *ptr = (uint8_t*)&req;
		ssize_t ret;
		ssize_t sizeread = 0;

		do {
			ret = read(ifd, ptr, sizeof(req)-sizeread);
			if(ret < 0) {
				fprintf(logfile, "Error reading, read returned %zd: %d %s\n", ret, errno, strerror(errno));
				exit(1);
			} else if (ret == 0) {
				fprintf(logfile, "Exiting ...\n");
				exit(0);
			} else {
				ptr += ret;
				sizeread += ret;
				fprintf(logfile, "Received %zd bytes of %zd\n", sizeread, sizeof(req));
			}
		} while(sizeread < sizeof(req));

		req.offset = ntohl(req.offset);
		req.length = ntohl(req.length);

		fprintf(logfile, "%s Received request:\n", serlogtime());
		fprintf(logfile, "Request: %x\n", req.command);
		fprintf(logfile, "ImageID: %x\n", req.imageid);
		fprintf(logfile, "Offset: %x\n", req.offset);
		fprintf(logfile, "Length: %x\n", req.length);

		struct response resp;
		uint8_t *buffer = NULL;

		switch(req.command) {
		case sCmdSize:
			resp.status = 0;
			resp.imageid = req.imageid;
			resp.length = (uint32_t)sb.st_size;
			resp.length = htonl(resp.length);
			ret = write(ofd, &resp, sizeof(resp));
			fprintf(logfile, "%s Wrote %zd bytes\n", serlogtime(), ret);
			break;
		case sCmdRead:
			resp.status = 0;
			resp.imageid = req.imageid;
			resp.length = req.length;
			resp.length = htonl(resp.length);
			ret = write(ofd, &resp, sizeof(resp));
			fprintf(logfile, "%s Wrote %zd bytes for read cmd\n", serlogtime(), ret);
			buffer = calloc(1, req.length);
			ret = pread(serfd, buffer, req.length, req.offset);
			fprintf(logfile, "%s Read %zd bytes from file\n", serlogtime(), ret);
			ret = write(ofd, buffer, req.length);
			fprintf(logfile, "%s Wrote %zd bytes from file\n", serlogtime(), ret);
			free(buffer);
			break;
		case sCmdWrite:
			buffer = malloc(req.length);
			fprintf(logfile, "%s Received write request, attempting to read %d bytes to write\n", serlogtime(), req.length);
			sizeread = 0;
			ptr = buffer;
			do {
				ret = read(ifd, ptr, req.length - sizeread);
				ptr += ret;
				sizeread += ret;
			}while(sizeread < req.length);

			fprintf(logfile, "%s Read %zd bytes of a requested %d\n", serlogtime(), sizeread, req.length);
			resp.status = 0;
			pwrite(serfd, buffer, req.length, req.offset);
			free(buffer);

			resp.imageid = req.imageid;
			resp.length = 0;
			ret = write(ofd, &resp, sizeof(resp));
			fprintf(logfile, "%s Wrote response size %zd\n", serlogtime(), ret);
			break;
		case sCmdEject:
			close(serfd);
			exit(0);
		};

	} while(1);
	
}
